'''
Generate Radial Distribution Functions (including GRID) from a pymatgen structure.

Various functions to compute pairwise atomic distances and bin
them into histograms, with optional Gaussian broadening.


The __name__ == '__main__' part in this module is used for test runs, 
so the input_file should contain a single structure
Batch calculation of the whole dataset, i.e. multiple input structures is done in the
__name__ == '__main__' part in data_explore.py

'''

import sys
import numpy as np
import argparse
import logging
import itertools
from tqdm import tqdm
from scipy.stats import gaussian_kde
from scipy.sparse import coo_matrix

# Handle pymatgen v2022 changes
try:
    from pymatgen import Structure
except ImportError:
    from pymatgen.core.structure import Structure   
try:
    from pymatgen import Lattice
except ImportError:
    from pymatgen.core.lattice import Lattice
from sklearn.neighbors import KernelDensity
from pyemd import emd

import warnings

def get_raw_rdf(structure, prim_cell_list, max_dist=10):
    '''
    Get pair distance in the structure at a given cutoff.
    
    This is the raw pair distance values before binning.    
    Currently the atomic species information is dropped.  

    Args:
        structure: pymatgen structure, typically a supercell
        max_dist: cutoff of the atomic pair distance
        prim_cell_list: index of the atoms of the selected primitive cell 
            (See NB1 in the header of this file)
    Return:
        A sorted 1d list of atomic pair distance 
    '''
    raw_rdf = []
    for site in prim_cell_list:
        for pair_site in structure.get_neighbors(site=structure[site], r=max_dist):
            raw_rdf.append(round(pair_site[1], 3))
    return sorted(raw_rdf)

def _sorted_neighbours(structure, max_dist):
    """
    Return the sorted neighbours of each site in a Structure.

    Parameters
    ----------
    structure : PyMatGen Structure

    max_dist: float
        Maximum cutoff distance for neighbour search in Angstroms

    Returns
    -------

    neighbours : list of list
        list of pymatgen PeriodicSite objects in distance order, of format
        [[site_0_neighbours], [site_1_neigbours], ...]

    """

    # PyMatGen returns all neighbours for all sites
    neighbours = structure.get_all_neighbors(max_dist)

    for site in neighbours:
        # Sort based on neighbour.nn_distance
        site.sort(key = lambda x: x.nn_distance)

    return neighbours

def _estimate_cutoff(structure, num_neighbours):
    """
    Estimate the cutoff required to achieve a certain number of neighbours based on atom density.

    Parameters
    ----------
    structure : PyMatGen structure
    num_neighbours : int
        Number of neighbours desired

    Returns
    -------
    cutoff : float
        Minimum suggested cutoff to achieve number of neighbours

    """

    atom_density = len(structure) / structure.volume

    # Cutoff based on number of atoms expected within sphere
    min_cutoff = ((3*num_neighbours) / (4*np.pi*atom_density))**(1./3)

    return min_cutoff * 1.1

def _estimate_neighbours(structure, cutoff):
    """
    Quickly estimate the minimum number of neighbours expected within a radius based on atom density.

    Parameters
    ----------
    structure : PyMatGen structure
    cutoff : float
        Cutoff distance in Angstrom

    Returns
    -------
    min_neighbours : int
        Minimum neighbours expected for a given cutoff

    See Also
    --------
    _estimate_cutoff

    """    

    atom_density = len(structure) / structure.volume
    return int(4/3*np.pi*cutoff**3 * atom_density)

def get_pairwise_distances(structure,
                           num_neighbours=None,
                           cutoff=None,
                           return_cutoff = False):
    """
    Return a sorted list of pairwise distances for a structure up to a cutoff.

    Parameters
    ----------

    structure : PyMatGen structure

    num_neighbours : int, default None
        Number of desired neighbours

    cutoff : float, default None
        Distance cutoff (in Ang) to calculate distance to

    NOTE: Only one of num_neighbours or cutoff must be supplied

    return_cutoff: bool, default False
        Return the distance cutoff used (usually helpful when num_neighbours is supplied)

    Returns
    -------

    neighbours : list of list
        list of pymatgen PeriodicSite objects in distance order, of format
        [[site_0_neighbours], [site_1_neigbours], ...]
    cutoff : float, optional
        distance cutoff used to generate neighbours

    Notes
    -----

    - If `num_neighbours` is defined, `return_cutoff` will give the cutoff required to return the same (minimum) number
      of neighbours. Please note that for structures with large numbers of sites, this may not yield the same number 
      of neighbours for all sites (use `num_neighbours` to give that behaviour).

    """

    assert (num_neighbours is None) ^ (cutoff is None), "You must specify only one of num_neighbours or cutoff."

    initial_cutoff = cutoff

    if num_neighbours:
        cutoff = _estimate_cutoff(structure, num_neighbours)
    
    complete = False
    while not complete:
        neighbours = _sorted_neighbours(structure, cutoff)
        if initial_cutoff:
            complete = True
        elif num_neighbours:
            min_count = min([len(i) for i in neighbours])
            if min_count < num_neighbours:
                cutoff *= 1.1
            else:
                complete = True

    # Ensure we are returning the correct number of neighbours
    if num_neighbours:
        for i, site in enumerate(neighbours):
            neighbours[i] = site[:num_neighbours]
        # Reset the cutoff to give the correct (minimum) number of neighbours across all sites
        cutoff = max([i[-1].nn_distance for i in neighbours]) + 0.0001
    
    if return_cutoff:
        return neighbours, cutoff
    else:
        return neighbours

def _neighbours_to_dists(neighbours):
    """
    Return the distances and occupancies from a Neighbour list.

    Parameters
    ----------

    neighbours : list of lists
        List of PyMatGen neighbour objects, ordered by site

    Returns
    -------

    dists : array-like
        Pairwise distances ordered as (sites, neighbours)
    occupancies : array-like
        Site occupancies ordered as (sites, neighbours)

    """

    simplified = []
    for site in neighbours:
        simplified.append([])
        for neig in site:
            simplified[-1].append([neig.nn_distance, neig.species.num_atoms])
    vals = np.array(simplified)

    # return distances and occupancies
    return vals[:,:,0], vals[:,:,1]

def __sample_dists_convolve(dists, weights, sampling_width, smearing):
    """ 
    Discretize distances and then convolve them with a Gaussian snearing.

    Bins the observed distances onto a fine grid, and then convolves the resulting
    histograms with a Gaussian broadening. The resulting fine grid is intended to be
    binned to a coarser distribution in order to approximate the integral

    Parameters
    ----------
    dists : array-like
        Array of pairwise distances.
    weights : array-like
        Weights for each distance (i.e. based on occupancies) with the same shape as dists
    sampling_width : float
        Width to use for fine sampling of smearing
    smearing : float
        Gaussian bandwidth to apply to distances

    Returns
    -------

    fine_bins : array-like
        Distances at which the Gaussian KDE has been sampled
    weights : array-like
        Weights for each distance (same shape as fine_bins)
    """
    
    from scipy.ndimage import gaussian_filter

    max_dist = dists.max()
    # Set up an array of finely spaced distances
    fine_bins = np.array([np.arange(0, max_dist+3*smearing+sampling_width, sampling_width)]*dists.shape[0])
    fine_counts = np.zeros((fine_bins.shape[0], fine_bins.shape[1]-1))

    for shell in np.arange(dists.shape[0]):
        fine_counts[shell,:] = np.histogram(dists[shell],
                                             bins=fine_bins[shell],
                                             weights=weights[shell],
                                             )[0]    

    # Calculate the filtered version. Note that scipy works in "pixel" units,
    # so the standard deviation is speified in terms of number of bins.
    # 
    # Should gaussian_filter take the std rather than std^2 of the Gaussian?
    smeared_wts = gaussian_filter(fine_counts, (0, smearing/sampling_width))

    return fine_bins[:, :-1], smeared_wts

def __sample_kde_fine(dists, weights, sampling_width, smearing):
    """ 
    Compute KDE estimate of broadening for each distance, and bin based on sampling_width.

    Parameters
    ----------
    dists : array-like
        Array of pairwise distances.
    weights : array-like
        Weights for each distance (i.e. based on occupancies) with the same shape as dists
    sampling_width : float
        Width to use for fine sampling of smearing
    smearing : float
        Gaussian bandwidth to apply to distances

    Returns
    -------

    fine_bins : array-like
        Distances at which the Gaussian KDE has been sampled
    weights : array-like
        Weights for each distance (same shape as fine_bins)

    Notes
    -----

    - We use scikit-learn's `KernelDensity` rather than scipy's `gaussian_kde` because the latter does not allow
      the bandwidth to be explicitly defined (instead, it is relative to the std of the data)
    
    """

    max_dist = dists.max()
    fine_bins = np.array([np.arange(0, max_dist+3*smearing, sampling_width)]*dists.shape[0])

    fine_counts = np.zeros_like(fine_bins)

    for shell in np.arange(dists.shape[0]):
        #kde = gaussian_kde(dists[shell], bw_method=smearing)
        #fine_counts[shell,:] = kde.evaluate(fine_bins[shell,:])
        kde = KernelDensity(bandwidth=smearing, kernel='gaussian')
        kde.fit(dists[shell].reshape([-1,1]), sample_weight=weights[shell])

        fine_counts[shell] = np.exp(kde.score_samples(fine_bins[shell,:].reshape(-1,1)))

    return fine_bins, fine_counts

def calculate_rdf(structure,
                  neighbours,
                  rdf_type='grid',
                  max_dist=None,
                  bin_width=0.1,
                  smearing = 0.1,
                  normed = True,
                  broadening_method = 'convolve',
                  return_sparse = False,
                  ):
    """
    Return the GRID (2D) RDF from a list of neighbours

    Parameters
    ----------

    structure : PyMatGen Structure
    neighbours : list
        List of neighbours generated using `get_pairwise_distances`
    rdf_type : str 
        Format of RDF to return
            'grid': return 2D RDF with distances sorted by neighbour order
            'simple': return the traditional 1D RDF
    max_dist : Float, default None
        Maximum cutoff to use when generating RDF representation. If None,
        the maximum observed distance will be used.
    bin_width : Float, default 0.1
        Width of histogram bins in Angstroms
    smearing : Float, default 0.1
        Width of Gaussian smearing to apply to distances prior to binning. If 0,
        distances will be binned without smearing
    normed : Bool, default True
        Whether to normalise the resulting histogram area(s). For 2D (grid) 
        histogram, each shell has area = 1
    broadening_method : str, default 'convolve'
        How to compute the Gaussian broadening (only applies if smearing > 0.)
        `convolve` : bin the distances into a finer distribution, and then apply a Gaussian
                     convolution to the results
        `kde`      : construct a KernelDensityEstimate based on the distances, and then 
                     compute the magnitude of this KDE on a fine distribution before binning the
                     results. 
    return_sparse : bool, default False
        If True, return a scipy coo_matrix format to save memory

    Notes
    -----

    **Broadening**

    Gaussian broadening is not implemented exactly due to the complexity of integrating a KDE
    across multiple bin ranges. Instead, two approaches are available which yield very similar overall
    results.
    broadening_method=`convolve` first bins the distances into a finely divided histogram (currently bin_width/5)
    before applying a Gaussian convolution across these bins. Formally this loses precision in the bond
    distances before broadening, but is otherwise correct (and fast).

    broadening_method=`kde` tries to compute a numerically correct KDE based on the exact distances, before then
    computing the value of this KDE on a fine scale (bin_width/5). The approximation is that the integral of 
    the KDE is assumed to be equal to the sum of these closely spaced KDE values. Although not strictly accurate, 
    this approximation is likely to affect every distance equally, therefore the overall histogram is probably 
    only different by a scale factor.

    The original `rdf_kde` method uses the same approach as broadening_method=`convolve`, with the exception that 
    the "fine" histogram was the same as the final histogram defined by `bin_width`. Please note that rounding was 
    also used, which may lead to slightly different numerical results.

    """

    dists, weights = _neighbours_to_dists(neighbours)
    # Arrange distances and weights into GRID-shell order
    dists = dists.T
    weights = weights.T

    # Multiply weights by central site occupancy
    weights = weights * [i.species.num_atoms for i in structure]

    # Set up vectors for distance and shell_number
    if max_dist:
        distance = np.arange(0,max_dist+bin_width, bin_width)
    else:
        max_dist = dists.max()
        distance = np.arange(0,max_dist+bin_width, bin_width)        

    shell_number = np.arange(dists.shape[0])

    if smearing > 0:
        # Calculate the gaussian broadening on a fine grid so that it can be binned later.
        # fine grid of bin_width/5 is somewhat arbitrary.
        convolve_scale = 5
        if broadening_method == 'kde':
            dists, weights = __sample_kde_fine(dists, weights, bin_width/convolve_scale, smearing)
        elif broadening_method == 'convolve':
            dists, weights = __sample_dists_convolve(dists, weights, bin_width/convolve_scale, smearing)

    if rdf_type == 'grid':
        binned_dists = np.zeros((shell_number.shape[0], distance.shape[0]-1))
        for shell in np.arange(dists.shape[0]):
            binned_dists[shell,:] = np.histogram(dists[shell],
                                                    bins=distance,
                                                    weights=weights[shell],
                                                    density=normed
                                                    )[0]

    elif rdf_type == 'simple':
        binned_dists = np.histogram(dists,
                                    bins=distance,
                                    weights = weights,
                                    density=normed
                                    )[0]
    else:
        raise ValueError(f"Unknown RDF type {rdf_type}")

    if return_sparse:
        binned_dists = coo_matrix(binned_dists)

    return binned_dists




def get_rdf_and_atoms(structure, prim_cell_list, max_dist=10):
    '''
    Get pair distance in the supercell, and the element symbols of the atom pair. 
    
    One atoms must be in the selected primtive cell.  
    The output dictionary should be like this:  
    {0: [[1.564, 'Si', 'O'],  # where '0' is the atom number
        [1.592, 'Si', 'O'],
        [1.735, 'Si', 'O'],
        [1.775, 'Si', 'O'],
        [2.924, 'Si', 'Si'],
        [3.128, 'Si', 'Si'],
        [3.148, 'Si', 'Si'], ...... } # list all pair with atom 0 within cutoff

    Args:
        structure: pymatgen structure, typically a supercell
        max_dist: cutoff of the atomic pair distance
        prim_cell_list: index of the atoms of the selected primitive cell
            (See NB1 in the header of this file)
    Return:
        A sortted list of atomic pair distance, with atom species
    '''

    warnings.warn('`get_rdf_and_atoms` will be deprecated in a future version. New code should use `get_pairwise_distances` instead', FutureWarning)

    rdf_atoms = {}
    for i, site in enumerate(prim_cell_list):
        rdf_atoms[i] = []
        site1 = structure[site].species_string
        for pair_site in structure.get_neighbors(site=structure[site], r=max_dist):
            site2 = pair_site[0].species_string
            rdf_atoms[i].append([round(pair_site[1],3), site1, site2])
        rdf_atoms[i].sort()
    return rdf_atoms


def rdf_histo(rdf_atoms, max_dist=10, bin_width=0.1):
    '''
    Convert the raw rdf with atoms to binned frequencies i.e. histogram.

    Args:
        rdf_atoms: pair distance of rdf with atomic speicies (output of get_rdf_and_atoms)
        max_dist: cutoff of the atomic pair distance
        bin_width: bin size for generating counts
    Return:
        Binned rdf frequencies for each shell of neasest neighbor
    '''

    warnings.warn('`rdf_histo` will be deprecated in a future version. New code should use `calculate_rdf` instead.', FutureWarning)

    # get the longest rdf number
    rdf_count = [ len(x) for x in rdf_atoms.values() ]
    rdf_len = np.array(rdf_count).max()

    # converse the rdf_atom into rdf in each shell, and only keep the distance values
    # e.g. rdf_nn_shell[0] contain all the pair distance of the first NN
    rdf_nn_shells = []
    for x in range(rdf_len):
         rdf_nn_shells.append( [line[x][0] 
                            for line in rdf_atoms.values() 
                            if len(line) > x] )

    bins = np.linspace(start=0, stop=max_dist, num=int(max_dist/bin_width)+1)
    # np.histogram also return the bin edge, which is not needed
    # so only the bin counts [0] is kept    
    rdf_bin = [ np.histogram(x, bins=bins, density=False)[0]
                for x in rdf_nn_shells ]
    return np.array(rdf_bin)

def find_all_neighbours(structure_list,
                        num_neighbours=None,
                        cutoff=None,
                        return_limits = True,
                        dryrun = False):
    """
    Return neighbour lists for an iterable list of structures.

    Parameters
    ----------
    
    structures : list of PyMatGen structures

    num_neighbours : int, default None
        Number of desired neighbours for each structure

    cutoff : float, default None
        Distance cutoff (in Ang) to find neighbours

    NOTE: Only one of num_neighbours or cutoff must be supplied

    return_limits : bool, default True
        If True, return either [min_neigh, max_neigh] or [min_cutoff, max_cutoff] depending on which parameter
        was supplied.

    dryrun : bool, default False
        If True, only estimate cutoffs for the given parameters, but don't calculate neighbours

    Returns
    -------

    neighbours : nested lists of neighbours
        list of pymatgen PeriodicSite objects in distance order, of format
        [[site_0_neighbours], [site_1_neigbours], ...]

    limits : tuple(min_limit, max_limit), optional
        If return_limits is True, the minimum and maximum limits are returned. The numbers depend on whether
        `cutoff` or `num_neighbours` was specified; for `cutoff` the smallest and largest number of neighbours
        are returned, while if `num_neighbours` was specified `limits` contains the minimum and maximum cutoff 
        distances.

    """

    assert (num_neighbours is None) ^ (cutoff is None), "You must specify only one of num_neighbours or cutoff."

    # Try to quickly estimate cutoffs or neighbours
    if num_neighbours:
        est_cutoffs = [_estimate_cutoff(i, num_neighbours) for i in structure_list]
        min_cut = min(est_cutoffs)
        max_cut = max(est_cutoffs)
        warnings.warn(f'The estimated cutoff distances range from {min_cut:.2f} to {max_cut:.2f} Angstroms', stacklevel=2)
    elif cutoff:
        est_neig = [_estimate_neighbours(i, cutoff) for i in structure_list]
        min_neig = min(est_neig)
        max_neig = max(est_neig)
        warnings.warn(f'The estimated number of neighbours ranges from {min_neig} to {max_neig}.', stacklevel=2)

    if dryrun:
        if return_limits:
            if num_neighbours:
                return None, (min_cut, max_cut)
            elif cutoff:
                return None, (min_neig, max_neig)
        else:
            return None


    neighbours = []
    neigh_range = [1E7,0]
    cutoff_range = [1E7,0]

    for struc in tqdm(structure_list, mininterval=5):
        neighbours.append(get_pairwise_distances(struc, num_neighbours=num_neighbours, cutoff=cutoff))
        if cutoff:
            current_neighbour_count = max([len(i) for i in neighbours[-1]])
            if current_neighbour_count < neigh_range[0]:
                neigh_range[0] = current_neighbour_count
            elif current_neighbour_count > neigh_range[1]:
                neigh_range[1] = current_neighbour_count
        elif num_neighbours:
            max_dist = neighbours[-1][-1][-1].nn_distance
            if max_dist < cutoff_range[0]:
                cutoff_range[0] = max_dist
            elif max_dist > cutoff_range[1]:
                cutoff_range[1] = max_dist

    if num_neighbours:
        warnings.warn(f'Computed cutoff ranges from {cutoff_range[0]:.4f} to {cutoff_range[1]:.4f}.', stacklevel=2)
    elif cutoff:
        warnings.warn(f'Computed number of neighbours ranges from {neigh_range[0]} to {neigh_range[1]}.', stacklevel=2)
    
    if return_limits:
        if num_neighbours:
            return neighbours, cutoff_range
        elif cutoff:
            return neighbours, neigh_range
    else:
        return neighbours

def rdf_stack_histo(rdf_atoms, structure, max_dist=10, bin_width=0.1, bond_direct=False):
    '''
    Convert the raw rdf with atoms to binned frequencies i.e. histogram
    and condsidering different atomic pairs

    Args:
        rdf_atoms: pair distance of rdf with atomic speicies (output of `get_rdf_and_atoms`)
        structure: pymatgen structure
        max_dist: cutoff of the atomic pair distance
        bin_width: bin size for generating counts
        bond_direct: if True, same atom pairs (e.g ['Si','O'] and ['O','Si']) are merged
    Return:
        Binned rdf frequencies for each shell of nearest neighbor
        and a string of ordered atomic pairs
    '''
    # get the longest rdf number
    rdf_count = [ len(x) for x in rdf_atoms.values() ]
    rdf_len = np.array(rdf_count).max()

    # converse the rdf_atom into rdf in each shell,
    # i.e. rdf_nn_shell[0] contain all the pair distance of the first NN
    rdf_nn_shells = []
    for x in range(rdf_len):
         rdf_nn_shells.append([ line[x] 
                            for line in rdf_atoms.values() 
                            if len(line) > x ])

    # breakdown each rdf_shell to atom pair dependent
    rdf_atom_pair_shells = []
    if bond_direct:
        # get all the atomic pairs
        atom_pair_list = list(itertools.product(structure.symbol_set, repeat=2))
        for rdf_shell in rdf_nn_shells:
            for atom_pair in atom_pair_list:
                rdf_atom_pair_shells.append([ x[0]
                                            for x in rdf_shell
                                            if x[1:] == list(atom_pair) ])
    else:
        atom_pair_list = list(itertools.combinations(structure.symbol_set, r=2)) \
                        + [ (a,a) for a in structure.symbol_set ]
        for rdf_shell in rdf_nn_shells:
            for atom_pair in atom_pair_list:
                rdf_atom_pair_shells.append([ x[0]
                                            for x in rdf_shell
                                            if (x[1:] == list(atom_pair) or 
                                                x[1:][::-1] == list(atom_pair)) ])    

    bins = np.linspace(start=0, stop=max_dist, num=int(max_dist/bin_width)+1)
    # np.histogram also return the bin edge, which is not needed
    # so only the bin counts [0] is kept    
    rdf_bin = [ np.histogram(x, bins=bins, density=False)[0]
                for x in rdf_atom_pair_shells ]
    return np.array(rdf_bin), atom_pair_list


def rdf_kde(rdf_atoms, max_dist=10, bin_width=0.1, bandwidth=0.1):
    '''
    Convert the raw rdf with atoms to binned frequencies with Gaussian smearing.

    Args:
        rdf_atoms: pair distance of rdf with atomic speicies (output of get_rdf_and_atoms)
        max_dist: cutoff of the atomic pair distance
        bin_width: bin size for generating counts
    Return:
        rdf_bin: Gaussian smeared rdf frequencies for each shell of neasest neighbor
    '''

    warnings.warn('`rdf_kde` will be deprecated in a future version. New code should use `calculate_rdf` instead.', FutureWarning)

    # get the longest rdf number
    rdf_count = [ len(x) for x in rdf_atoms.values() ]
    rdf_len = np.array(rdf_count).max()

    # converse the rdf_atom into rdf in each shell,
    # and only keep the distance values
    # e.g. rdf_nn_shell[0] contain all the pair distance of the first NN
    rdf_nn_shells = []
    for x in range(rdf_len):
         rdf_nn_shells.append( [line[x][0]
                            for line in rdf_atoms.values() 
                            if len(line) > x] )

    # the kernel density method need a 2d input, so add a new axis
    bins = np.linspace(start=0, stop=max_dist, num=int(max_dist/bin_width)+1)[:, np.newaxis]

    rdf_bin = []
    kde = KernelDensity(kernel='gaussian', bandwidth=bandwidth)
    for x in rdf_nn_shells:
        log_dens = kde.fit(np.array(x)[:, np.newaxis]).score_samples(bins)
        # Due to Gaussian smearing hitting the edge of the RDF range,
        # occasionally the densities do not sum to 1.
        # We normalize here, so Wasserstein_distance doesn't need to
        dens = np.exp(log_dens)
        dens = dens / dens.sum()
        rdf_bin.append(dens)

    return np.array(rdf_bin)


# def shell_similarity(rdf_bin):
#     '''
#     Calculate the earth mover distance (EMD) between adjacent rdf shells  
#     i.e. the first value is the EMD between the first shell and second

#     Args:
#         rdf_bin: calculated rdf, only rdf from function rdf_histo has been tested
#     Return:
#         np array of the similarity, with length (len(rdf_bin)-1)
#     '''
#     shell_dist = np.zeros((len(rdf_bin), len(rdf_bin)))
#     dist_matrix = np.ones((len(rdf_bin[0]), len(rdf_bin[0])))
#     np.fill_diagonal(dist_matrix, 0)
#     for i, r1 in enumerate(rdf_bin):
#         for j, r2 in enumerate(rdf_bin):
#             if i < j:
#                 dissim = emd(r1.astype('float64'), r2.astype('float64'), 
#                         dist_matrix.astype('float64'))
#                 shell_dist[i,j] = dissim
#                 shell_dist[j,i] = dissim
#    return shell_dist


# if __name__ == '__main__':
#     parser = argparse.ArgumentParser(description='Calculate RDF with atoms',
#                                     formatter_class=argparse.RawTextHelpFormatter)
#     parser.add_argument('--input_file', type=str, default=None,
#                         help='Input CIF containing the crystal structure')
#     parser.add_argument('--task', type=str, default='shell_similarity',
#                         help='what to be calculated: \n' +
#                             '   rdf: calculate RDF \n' +
#                             '   stack_rdf: RDF with different atomic pair \n' +
#                             '   shell_similarity: the similarity between two nearest shell \n' +
#                             '   raw_rdf: origin 1D RDF as sorted list'
#                       )
#     parser.add_argument('--output', type=str, default=None,
#                         help='Output RDF')
#     parser.add_argument('--max_dist', type=float, default=10.0,
#                         help='Cutoff distance of the RDF')
#     parser.add_argument('--trim', type=int, default=30,
#                         help='the number of shells for RDF, 0 means no trim')

#     args = parser.parse_args()
#     input_file = args.input_file
#     task = args.task
#     outfile = args.output
#     max_dist = args.max_dist
#     trim = args.trim

#     np.set_printoptions(threshold=sys.maxsize) # print the whole array

#     if input_file:
#         # read a structure from cif to a pymatgen structure
#         struct = Structure.from_file(filename=input_file, primitive=True)
#     else:
#         # if a input structure is not provide, the code is in test mode
#         # and nacl structure will be used for test propose
#         nacl = Structure.from_spacegroup('Fm-3m', Lattice.cubic(5.6), 
#                                         ['Na', 'Cl'], [[0.5, 0.5, 0.5], [0, 0, 0]])
#         struct = nacl.get_primitive_structure()
    
#     # The 'prim_cell_list' is used with the 'extend_structure' function, when the function
#     # is deprecated, this variable is kept maybe useful in the future
#     prim_cell_list = list(range(len(struct)))

#     if task == 'rdf':
#         rdf_atoms = get_rdf_and_atoms(structure=struct, prim_cell_list=prim_cell_list, 
#                                         max_dist=max_dist)
#         rdf_bin = rdf_histo(rdf_atoms=rdf_atoms, max_dist=max_dist, bin_width=0.1)
#         if outfile:
#             np.savetxt(outfile, rdf_bin.transpose(), delimiter=' ',fmt='%i')
    
#     elif task == 'stack_rdf':
#         rdf_atoms = get_rdf_and_atoms(structure=struct, prim_cell_list=prim_cell_list, 
#                                         max_dist=max_dist)
#         rdf_bin, atom_pairs = rdf_stack_histo(rdf_atoms=rdf_atoms, structure=struct, 
#                                             max_dist=max_dist, bin_width=0.1)
#         if outfile:
#             print(atom_pairs)
#             # transpose the ndarray for import into the plot program
#             np.savetxt(outfile, rdf_bin.transpose(), delimiter=' ',fmt='%i')
    
#     # elif task == 'shell_similarity':
#     #     rdf_atoms = get_rdf_and_atoms(structure=struct, prim_cell_list=prim_cell_list, 
#     #                                     max_dist=max_dist)
#     #     rdf_bin = rdf_histo(rdf_atoms=rdf_atoms, max_dist=max_dist, bin_width=0.1)
#     #     if trim != 0:
#     #         rdf_bin = rdf_bin[:trim]

#     #     shell_simi = shell_similarity(rdf_bin)
#     #     print(shell_simi)
#     #     if outfile:
#     #         np.savetxt(outfile, shell_simi, delimiter=' ', fmt='%.3f')
    
#     elif task == 'raw_rdf':
#         raw_rdf = get_raw_rdf(structure=struct, prim_cell_list=prim_cell_list, max_dist=max_dist)
#     else:
#         print('This task is not supported')




'''
Visualization results of machine learning models.
'''

import json
import heapq
import numpy as np
import pandas as pd
from scipy.stats import wasserstein_distance
from sklearn.model_selection import train_test_split
from pyemd import emd_with_flow


def calc_obs_vs_pred(funct, X_data, y_data, test_size, outdir='../'):
    '''
    The observation vs prediction plot
    '''
    X_train, X_test, y_train, y_test = \
            train_test_split(X_data, y_data, test_size=test_size, random_state=1)
    funct.fit(X_train, y_train)
    y_pred = funct.predict(X_test)
    y_pred_train = funct.predict(X_train)
    np.savetxt(outdir + 'test.' + str(test_size), 
                np.stack([y_test, y_pred]).transpose(), 
                delimiter=' ', fmt='%.3f')
    np.savetxt(outdir + 'train.' + str(test_size), 
                np.stack([y_train, y_pred_train]).transpose(), 
                delimiter=' ', fmt='%.3f')


def binarize_output(y_test, y_pred, threshold=None, nelem=None, save_to_file=False):
    '''
    Make the multilabe output decimals into 0 or 1
    according to the number of elements or a threshold

    Args:
        y_test: the ground truth label of the test data
        y_pred: prediciton of the test data
        threshold: above this value the output will be set to 1 and others 0
            If 'None', use the number (n) of non-zero labels in the ground
            truth, i.e. the top n largest values are set to 1 and others 0
        nelem: if None, use the number elements in the ground truth
            or given as an integer value
        save_to_file:
    Return:
        y_bin: a numpy array of binarized prediction
    '''
    y_bin = []
    nlabel = len(y_pred[0])
    for i, y in enumerate(y_pred):
        if threshold:
            y_new = np.zeros(nlabel)
            y_new[np.where(y > threshold)] = 1
        else:
            # the ground truth is 0 or 1, so the sum gives number of elements
            # note that sum in on the y_test i.e. ground truth
            if nelem:
                n_elem = nelem
            else:
                n_elem = y_test[i].sum()

            indice = y.argsort()[-n_elem:]
            y_new = np.zeros(nlabel)
            y_new[indice] = 1

        y_bin.append(y_new)

    if save_to_file:
        # save the ground truth of the test data
        # the current workdir is usually the one has the RDF files
        # so ../ means the files will be saved in parent folder
        np.savetxt('../multilabel_' + str(threshold) + '_test', 
                    y_test, delimiter=' ', fmt='%1i')
        # save the prediction values of the test data
        np.savetxt('../multilabel_' + str(threshold) + '_pred', 
                    y_pred, delimiter=' ', fmt='%.3f')
        # save the rounded values
        np.savetxt('../multilabel_' + str(threshold) + '_bin', 
                    np.stack(y_bin), delimiter=' ', fmt='%1i')

    return np.stack(y_bin)


def n_best_and_worst(y_test, y_pred, metrics_values, n_visual=100, 
                    method='confusion_matrix'):
    '''
    Output the n best and n worst prediction for visualization purpose
    This use heapq module to get n largest and smallest metrics values  
    For that has n middle values, see function 'n_best_middle_worst'

    Args:
        y_test: test data   
        y_pred: prediction data
        metrics_values: the criterion to select n best and worst
        n_visual: number of best and worst predictions 
        method: how the y vector is visualized
            original: original vectors, typically 78 elements
            confusion_matrix: 
    Return:

    '''
    # Give both index and value using heapq
    n_large = heapq.nlargest(n_visual, enumerate(metrics_values), key=lambda x: x[1])
    n_small = heapq.nsmallest(n_visual, enumerate(metrics_values), key=lambda x: x[1])

    large_samples = []
    small_samples = []
    middle_samples = []
    if method == 'original':
        for i, m_value in n_large:
            large_samples.append(y_test[i])
            large_samples.append(y_pred[i])
        for i, m_value in n_small:
            small_samples.append(y_test[i])
            small_samples.append(y_pred[i])
        return np.stack(large_samples).transpose(), np.stack(small_samples).transpose()
    elif method == 'confusion_matrix':
        for i, m_value in n_large:
            # confusion matrix, 0: true negetive,  1: false negative, 
            #                   2: false positive, 3: true positive
            large_samples.append(y_pred[i] * 2 + y_test[i])    
        for i, m_value in n_small:
            small_samples.append(y_pred[i] * 2 + y_test[i])
        return np.stack(large_samples).transpose(), np.stack(small_samples).transpose()
    else:
        print('This method is not supported in n_best_and_worst')


def n_best_middle_worst(y_test, y_pred, metrics_values, n_visual=100, 
                    method='confusion_matrix'):
    '''
    Output the n best, n middle, and n worst prediction for visualization purpose

    Args:
        y_test: test data   
        y_pred: prediction data
        metrics_values: the criterion to select n best and worst
        n_visual: number of best and worst predictions 
        method: how the y vector is visualized
            original: original vectors, typically 78 elements
            confusion_matrix: 
    Return:

    '''
    m_values = pd.DataFrame(metrics_values).sort_values(0)
    middle_index = int(len(metrics_values) / 2)
    middle_range = [ middle_index - int(n_visual / 2), middle_index + int(n_visual / 2)]

    large_samples = []
    small_samples = []
    middle_samples = []
    if method == 'original':
        for i in m_values[-n_visual:].index.values:
            large_samples.append(y_test[i])
            large_samples.append(y_pred[i])
        for i in m_values[middle_range[0]:middle_range[1]].index.values:
            middle_samples.append(y_test[i])
            middle_samples.append(y_pred[i])
        for i in m_values[:n_visual].index.values:
            small_samples.append(y_test[i])
            small_samples.append(y_pred[i])
        return np.stack(large_samples).transpose(), \
                np.stack(middle_samples).transpose(), \
                np.stack(small_samples).transpose()
    
    elif method == 'confusion_matrix':
        for i in m_values[-n_visual:].index.values:
            # confusion matrix, 0: true negetive,  1: false negative, 
            #                   2: false positive, 3: true positive
            large_samples.append(y_pred[i] * 2 + y_test[i])    
        for i in m_values[middle_range[0]:middle_range[1]].index.values:
            middle_samples.append(y_pred[i] * 2 + y_test[i])
        for i in m_values[:n_visual].index.values:
            small_samples.append(y_pred[i] * 2 + y_test[i])
        return np.stack(large_samples).transpose(), \
                np.stack(middle_samples).transpose(), \
                np.stack(small_samples).transpose()    

    else:
        print('This method is not supported in n_best_and_worst')


if __name__ == '__main__':
    pettifor = ['Cs', 'Rb', 'K', 'Na', 'Li', 'Ba', 'Sr', 'Ca', 'Yb', 'Eu', 'Y',  'Sc', 'Lu', 'Tm', 'Er', 'Ho', 
    'Dy', 'Tb', 'Gd', 'Sm', 'Pm', 'Nd', 'Pr', 'Ce', 'La', 'Zr', 'Hf', 'Ti', 'Nb', 'Ta', 'V',  'Mo', 
    'W',  'Cr', 'Tc', 'Re', 'Mn', 'Fe', 'Os', 'Ru', 'Co', 'Ir', 'Rh', 'Ni', 'Pt', 'Pd', 'Au', 'Ag', 
    'Cu', 'Mg', 'Hg', 'Cd', 'Zn', 'Be', 'Tl', 'In', 'Al', 'Ga', 'Pb', 'Sn', 'Ge', 'Si', 'B',  'Bi', 
    'Sb', 'As', 'P',  'Te', 'Se', 'S', 'C', 'I', 'Br', 'Cl', 'N', 'O', 'F', 'H']




"""
Interface to collate or generate input data.

Helper functions to simplify extraction of data from the Materials Project
or to generate 'standard' datasets (such as artificial perovskites).
"""

# Bash script to generate RDF and GRID for structures
'''
for rdf in smooth_rdf vanilla_rdf ; do cd $rdf ; \
for file in pero_distortion pero_lattice rp_srtio3 ; \
do python ../../../descriptors/data_explore.py --input_file ../$file.json \
 --task rdf_similarity  --output_file ../${file}_$rdf ; \
done ; cd .. ; done 
'''

import json
import math
import argparse
import time
import numpy as np
import pandas as pd
import os
import warnings

import tqdm
from collections import Counter
try:
    from pymatgen import Structure
except ImportError:
    from pymatgen.core.structure import Structure   
try:
    from pymatgen import Lattice
except ImportError:
    from pymatgen.core.lattice import Lattice

from pymatgen.ext.matproj import MPRester

from . import extendRDF
from gridrdf.composition import elements_selection
from gridrdf.otherRDFs import origin_rdf_histo
from gridrdf.data_io import rdf_read, rdf_read_parallel


def nacl():
    '''
    Generate NaCl structure, usually for test purpose

    Args:
        None
    Return:
        a pymatgen structure of NaCl
    '''
    nacl = Structure.from_spacegroup(
        'Fm-3m', Lattice.cubic(5.6),['Na', 'Cl'],[[0.5, 0.5, 0.5], [0, 0, 0]] )
    return nacl


def get_MP_bulk_modulus_data(APIkey): 
    '''
    Get all Material project structures with calculated elastic properties

    Args:
        APIkey: a string 'Hecxxxxxxxxxxx' needed to be applied from MP website
    Return:
        A json object containing
            MP task id
            Bulk modulus
            Shear modulus
            Elastic anisotropy
            CIF of structure
    '''
    # put personal API key here
    q = MPRester(APIkey)   
    print(f"Using Materials Project database v{q.get_database_version()}")
    
    d = q.query(criteria={'elasticity.K_VRH':{'$nin': [None, '']}}, 
                properties=['task_id', 
                            'elasticity.K_VRH', 
                            'elasticity.G_VRH', 
                            'elasticity.elastic_anisotropy', 
                            'cif'])
    
    
    return d


def get_ICSD_CIFs_from_MP(APIkey): 
    '''
    Get all ICSD structures in the Material project 

    Args:
        APIkey: a string 'Hecxxxxxxxxxxx' needed to be applied from MP website
    Return:
        A json object containing
            MP task id
            ICSD id
            CIF of structure
    '''
    q = MPRester(APIkey)   
    d = q.query(criteria={'icsd_ids':{'$nin': [None, []]}}, 
                properties=['task_id', 'icsd_ids', 'cif'])
    return d

def _insert_field(infile1='num_shell', infile2='MP_modulus_all.json', 
                outfile='MP_modulus_v4.json'):
    '''
    Insert new file in the json file

    Args:
        infile1: file containing values to be inserted, the field should
            consult data_explore.py
        infile2: file into which new field will be inserted
        outfile: a new file contains new inserted fields
    Return:
        None
    '''
    results = np.loadtxt(infile1, delimiter=' ')
    with open(infile2, 'r') as f:
        data = json.load(f)

    for i, d in enumerate(data):
        d['average_bond_length'] = results[i][2]
        d['bond_length_std'] = results[i][3]    

    with open(outfile, 'w') as f:
        json.dump(data, f, indent=1)
    
    return


def _json_order():
    '''
    Make the perovskite structure in the order of lattice constant
    '''

    with open('pero2.json','r') as f:
        data = json.load(f)

    l = []
    for d in data:
        struct = Structure.from_str(d['cif'], fmt='cif')
        l.append(struct.lattice.a)

    df2 = pd.DataFrame(l, columns=['lattice'])
    new_index = df2.sort_values('lattice').index.values

    d2 = []
    for i in new_index:
        d2.append(data[i])

    with open('pero3.json','w') as f:
        json.dump(d2,f,indent=1)

def make_distorted_perovskite(outfile = None):
    '''
    used for testing purpose for EMD + extended RDF as a similarity measure

    '''
    cif = "# generated using pymatgen\ndata_SrTiO3\n_symmetry_space_group_name_H-M   'P 1'\n_cell_length_a   3.94513000\n_cell_length_b   3.94513000\n_cell_length_c   3.94513000\n_cell_angle_alpha   90.00000000\n_cell_angle_beta   90.00000000\n_cell_angle_gamma   90.00000000\n_symmetry_Int_Tables_number   1\n_chemical_formula_structural   SrTiO3\n_chemical_formula_sum   'Sr1 Ti1 O3'\n_cell_volume   61.40220340\n_cell_formula_units_Z   1\nloop_\n _symmetry_equiv_pos_site_id\n _symmetry_equiv_pos_as_xyz\n  1  'x, y, z'\nloop_\n _atom_site_type_symbol\n _atom_site_label\n _atom_site_symmetry_multiplicity\n _atom_site_fract_x\n _atom_site_fract_y\n _atom_site_fract_z\n _atom_site_occupancy\n  Sr  Sr0  1  0.00000000  0.00000000  0.00000000  1\n  Ti  Ti1  1  0.50000000  0.50000000  0.50000000  1\n  O  O2  1  0.50000000  0.00000000  0.50000000  1\n  O  O3  1  0.50000000  0.50000000  0.00000000  1\n  O  O4  1  0.00000000  0.50000000  0.50000000  1\n"
    s = Structure.from_str(cif, fmt='cif')

    posits = [
            [0.50, 0.50, 0.50],
            [0.50, 0.50, 0.51],
            [0.50, 0.50, 0.52],
            [0.50, 0.50, 0.53],
            [0.50, 0.507, 0.507],
            [0.506, 0.506, 0.506]
        ]

    all_dict = []
    for i, posit in enumerate(posits):
        one_dict = {}
        one_dict['task_id'] = 'pero_distort_' + str(i)
        s[1] = 'Ti', posit
        one_dict['cif'] = s.to(fmt='cif')
        all_dict.append(one_dict)

    if outfile is not None:
        with open(outfile,'w') as f:
            json.dump(all_dict, f, indent=1)

    return all_dict

def perovskite_different_lattice(outfile=None):
    cif = "# generated using pymatgen\ndata_SrTiO3\n_symmetry_space_group_name_H-M   'P 1'\n_cell_length_a   3.94513000\n_cell_length_b   3.94513000\n_cell_length_c   3.94513000\n_cell_angle_alpha   90.00000000\n_cell_angle_beta   90.00000000\n_cell_angle_gamma   90.00000000\n_symmetry_Int_Tables_number   1\n_chemical_formula_structural   SrTiO3\n_chemical_formula_sum   'Sr1 Ti1 O3'\n_cell_volume   61.40220340\n_cell_formula_units_Z   1\nloop_\n _symmetry_equiv_pos_site_id\n _symmetry_equiv_pos_as_xyz\n  1  'x, y, z'\nloop_\n _atom_site_type_symbol\n _atom_site_label\n _atom_site_symmetry_multiplicity\n _atom_site_fract_x\n _atom_site_fract_y\n _atom_site_fract_z\n _atom_site_occupancy\n  Sr  Sr0  1  0.00000000  0.00000000  0.00000000  1\n  Ti  Ti1  1  0.50000000  0.50000000  0.50000000  1\n  O  O2  1  0.50000000  0.00000000  0.50000000  1\n  O  O3  1  0.50000000  0.50000000  0.00000000  1\n  O  O4  1  0.00000000  0.50000000  0.50000000  1\n"
    s = Structure.from_str(cif, fmt='cif')

    all_dict = []
    for lat in np.linspace(3, 6, 61):
        one_dict = {}
        one_dict['task_id'] = 'pero_latt_' + str(round(lat,3))
        new_latt = Lattice.from_parameters(lat, lat, lat, 90, 90, 90)
        s.lattice = new_latt
        one_dict['cif'] = s.to(fmt='cif')
        all_dict.append(one_dict)

    if outfile is not None:
        with open(outfile,'w') as f:
            json.dump(all_dict, f, indent=1)
            
    return all_dict
            
def batch_rdf(data,
              max_dist=10,
              num_neighbours = 100,
              bin_width=0.1,
              broadening=0.1,
              normalize=True,
              output_dir = './',
              ):
    '''
    Read structures and output the GRID representations to files

    Parameters
    ----------

    data: iterable
        List of dicts containing (as a minimum) the 'task_id' and 'cif' keys. Intended to 
        be generated from Materials Project.
    max_dist: float, optional
        Maximum distance to compute shells to. Alternatively, specify "num_neighbours" directly 
        to avoid truncating shells later.
    num_neighbours : int, optional
        Number of nearest neighbours to consider in RDF calculation (alternatively, specify max_dist).
    broadening : float, default 0.1
        Gaussian broadening to apply to RDF representation. If 0, histogram will be 
        created without broadening.
        For more control over the broadening process, see `extendRDF.calculate_RDF`.
    normalize : Bool, default True
        Whether to normalize the resulting histogram area.

        
    Notes
    -----

    If both max_dist and num_neighbours are specified, GRID will be calculated on the basis of 
    num_neighbours.
    '''

    structures = []
    # First, generate PyMatGen objects from CIF representations
    for d in data:
        structures.append(Structure.from_str(d['cif'], fmt='cif'))

    orig_max_dist = max_dist
    # Determine whether to use max_dist or num_neighbours
    if num_neighbours and max_dist:
        max_dist = None

    # Generate all neighbour lists for all structures to the same cutoff
    neighbours, cutoffs = extendRDF.find_all_neighbours(structures,
                                               num_neighbours=num_neighbours,
                                               cutoff=max_dist,
                                               return_limits = True,
                                               dryrun = False)
    
    max_dist = max(np.round(cutoffs[1], 1), orig_max_dist)
    if max_dist != orig_max_dist:
        print(f"Maximum distance has been updated to {max_dist} to account for {num_neighbours} neighbours")


    rdf_results = []

    # Compute RDF representation
    for i, struct in tqdm.tqdm(enumerate(structures)):
        rdf_results.append(extendRDF.calculate_rdf(struct,
                                                   neighbours[i],
                                                   rdf_type='grid',
                                                   max_dist=max_dist,
                                                   bin_width=bin_width,
                                                   smearing = broadening,
                                                   normed = normalize,
                                                   broadening_method = 'convolve',
                                                   return_sparse = False)
                            )

    if not os.path.isdir(output_dir):
        os.mkdir(output_dir)

    # Save GRIDs to individual output files.
    warnings.warn("GRIDs will be saved in a different format in a future version.", FutureWarning)
    for i, GRID in enumerate(rdf_results):
        outfile = os.path.normpath(os.path.join(output_dir, data[i]['task_id']))
        np.savetxt(outfile, GRID, delimiter=' ', fmt='%.6f')
    
def trim_rdf_bins(data,
                  all_rdf,
                  number_of_shells,
                  write_to_disk = True,
                  output_dir = './',
                  disk_time = 0.01,
                  ):
    """ Truncate all RDFs to the largest number of bins possible in the group. 
    
    Args:
        data: input data from Materials Project
        all_rdf: list of np.arrays containing RDF (or GRID) information
        number_of_shells: number of RDF shells to trim all_GRID entries to (uses
                          first RDF axis by default)
        write_to_disk: whether to update existing RDF files on disk based on results
                       (default False). This will update GRIDs that are trimmed, but 
                       does NOT remove GRIDs that are shorter than the required length.
        outdir: directory to write (or update) RDF files to
        gzip_file: in case RDF files take too much disk space, set 
                    this to true to gzip the files (not yet test)
        disk_time: time to allow between disk writes (default 0.01)
        
    Returns:
        num_trimmed: The number of RDF entries that are above the cutoff length and were
                     trimmed to match the cutoff.
        num_removed: The number of RDF entries that were below the cutoff and were removed
                     from `data`.
    """
    
    assert number_of_shells > 0
    max_shells = np.max([i.shape[0] for i in all_rdf])
    assert number_of_shells <= max_shells, f"Number of shells chosen ({number_of_shells}) is bigger than the largest number calculated ({max_shells})"
    
    num_removed = 0
    num_trimmed = 0
    new_data = []
    new_rdfs = []
    
    for i, d in enumerate(data):
        if all_rdf[i].shape[0] < number_of_shells:
            num_removed += 1
            continue
                
        elif all_rdf[i].shape[0] > number_of_shells:
            #print('Started: ', all_rdf[i].shape[0])
            all_rdf[i] = all_rdf[i][:number_of_shells]
            num_trimmed += 1
            #print('Ended: ', all_rdf[i].shape[0])
            
            if write_to_disk:
                outfile = os.path.normpath(os.path.join(output_dir, d['task_id']))

                np.savetxt(outfile, all_rdf[i][:number_of_shells], delimiter=' ', fmt='%.6f')
                time.sleep(disk_time)
            
                
        new_data.append(data[i])
        new_rdfs.append(all_rdf[i])
                
    return new_data, new_rdfs, num_trimmed, num_removed
    
    
def main(data_source = 'MP_modulus.json',
         tasks = ['grid_rdf_kde',],
         MP_API_KEY = None,
         composition = {'elem':[], 'type':'consist'},
         output_dir = './outputs/',
         output_file = None,
         space_groups = [],
         data_property = (None, -np.inf, np.inf),
         num_grid_shells = 100,
         max_dist = 10,
         GRIDS = None,
        ):
    """ Main logic of data preparation. 
    
    
    Args:
        data_source: str
            Source of MP data to process. Possible options are:
                - nacl
                    - data are generated using `nacl` function
                - perovskite_lattice
                    - generated using `perovskite_different_lattice`
                - perovskite_distort
                    - generated using `make_distorted_perovskite`
                - None or MP_bulk_modulus
                    - data containing bulk moduli are extracted from MP.
                - ends with '.json'
                    - data are assumed to be in a json-formatted file
        tasks: list of str
            One or more tasks to be performed based on the input data in order supplied.
            Options are:
                - subset_composition
                    Filter based on elements. See required 'composition' argument.
                - subset_grid_len
                    Filter any structures containing fewer than `min_grid_groups`
                    GRID groups for the given distance cutoff (requires GRID files 
                    to have been generated first, either as a separate task or by 
                    specifying a suitable `output_dir`).    
                - subset_space_group
                    Filter structures by spacegroup number(s) as a list
                - subset_property
                    Filter by materials having a specific property (defined within the data source)
                - grid_rdf_bin
                    Compute GRID with basic histogram binning
                - grid_rdf_kde
                    Compute GRID with Gaussian smoothing of distances
                - original_rdf
                    Compute basic RDF
        MP_API_KEY: str
            Materials project API key 
        composition: dict
            If task == subset_composition, this dict is required. Keys are 
                'elem': list of species to be passed to `composition.elements_selection`
                'type': method to filter structures: 'include' (greedy AND), 'exclude' (NOT)
                        or 'consist' (ONLY)
        output_dir : str
            Location to save outputted file(s), for instance GRIDs.
        output_file : str
            File name for filtered data (if required).
        space_groups : list of ints
            If task == 'subset_space_group', this argument is the space group numbers
            to be retained. If len(space_groups) == 0, all space groups will be kept.
        data_property : list
            Property contained in data_source (e.g. 'elasticity.K_VRH') and the (min, max) 
            values it should adopt (inclusive)
        max_dist : float
            Maximum distance to calculate RDFs up to.
        num_grid_shells : int
            Number of GRID groups to trim/filter data to
                        
            
                    
    """

    assert 'type' in composition
    assert 'elem' in composition
    
    if data_source.lower() == 'nacl':
        data = nacl()
    elif data_source.lower() == 'perovskite_lattice':
        data = perovskite_different_lattice()
    elif data_source.lower() == 'perovskite_distort':
        data = make_distorted_perovskite()
    elif data_source.endswith('.json'):
        with open(data_source, 'r') as f:
            data = json.load(f)
    elif data_source is None or data_source.lower() == 'mp_bulk_modulus':
        if MP_API_KEY is None:
            raise ValueError('To access the materials project API, an API key is required')
        data = get_MP_bulk_modulus_data(MP_API_KEY)
    else:
        raise ValueError(f'Unknown data source {data_source}')
        
    original_length = len(data)
        
    
    for task in tasks:  
        print("Performing ", task, "on ", len(data), "entries.")
        if task == 'subset_composition':
            #print(elem_list)
            data = elements_selection(data, elem_list=composition['elem'], mode=composition['type'])

                
        elif task == 'subset_space_group':
            if len(space_groups) == 0:
                sg_kept = range(230, 0, -1)
            else:
                sg_kept = [int(i) for i in space_groups]
 
            for d in data:
                struct = Structure.from_str(d['cif'], fmt='cif')
                sg_num = struct.get_space_group_info()[1]
                if sg_num not in sg_kept:
                    data.remove(d)

        elif task == 'subset_grid_len':
            if GRIDS is None:
                all_rdf = rdf_read_parallel(data, output_dir)
            else:
                all_rdf = GRIDS
                
            data, all_rdf, trimmed, removed = trim_rdf_bins(data = data,
                                                             all_rdf = all_rdf,
                                                             number_of_shells = num_grid_shells,
                                                             write_to_disk = True,
                                                             output_dir = output_dir,
                                                             )
                                                             
            #GRIDS = all_rdf
            
            print(f"    Removed {removed} entries and trimmed {trimmed} entries based on a cutoff of {num_grid_shells} shells")


        elif task == 'subset_property':
            key = data_property[0]
            key_min = data_property[1]
            key_max = data_property[2]
            for d in data[:]:
                if d[key] < key_min or d[key] > key_max:
                    data.remove(d)
                    
        elif task == 'grid_rdf_bin':
            # Compute un-broadened GRID
            batch_rdf(data, max_dist=max_dist, num_neighbours = num_grid_shells, broadening=0.0, output_dir = output_dir)
        elif task == 'grid_rdf_kde':
            #Compute 0.1Ang broadened GRID
            batch_rdf(data, max_dist=max_dist, num_neighbours = num_grid_shells, broadening=0.1, output_dir = output_dir)
        elif task == 'original_rdf':
            origin_rdf_histo(data, max_dist=max_dist, output_dir = output_dir)

        else:
            raise ValueError(f'Unknown task {task}')
            
        print("Removed {} structure entries based on {}".format(original_length - len(data), task))
        original_length = len(data)

        
        
    # Save subset to new output file if needed
    if output_file is not None:
        outf = os.path.normpath(os.path.join(output_dir, output_file))
        if not os.path.isdir(output_dir):
            os.mkdir(output_dir)
        with open(outf, 'w') as f:
            json.dump(data, f, indent=1)
            
    return data

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Dataset prepare',
                                    formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('--data_source', type=str, default='../MP_modulus.json',
                        help='the source of structure data in materials project format, as json file.')
    parser.add_argument('-f', '--output_file', type=str, default='subset.json',
                        help='outpu')
    parser.add_argument('-d', '--output_dir', type=str, default='./',
                        help='Directory to output data file(s) to')
    parser.add_argument('--tasks', nargs='+', type=str, default='',
                        help='Which task(s) to perform on the dataset: \n' +
                        
                            '   subset_composition: select a subset which have specified elements: \n' +
                            '   subset_grid_len: drop all the ext-RDF with a length less than 100 \n' +
                            '   subset_space_group: restrict to specified spacegroup number(s)\n' +
                            '   grid_rdf_bin: compute GRID for all data_sources with simple binning\n' +
                            '   grid_rdf_kde: compute GRID for all data_sources with Gaussian distance broadening \n' + 
                            '   original_rdf: compute basic RDF for all data_sources'
                      )
                      
    parser.add_argument('--elem_list', nargs = '+',     type=str, default='O',
                        help='only used for subset task')
    parser.add_argument('--elem_method', type=str, default='include',
                        help='method to use for composition filtering (include, exclude or consist)')
    parser.add_argument('--spacegroups', nargs='+', type=str, default='',
                        help='spacegroup numbers to filter by (can include inclusive ranges, i.e. 40-56)')
    parser.add_argument('--max_dist', type=float, default=10.0,
                        help='Cutoff distance of the RDF')       
    parser.add_argument('--min_grid_groups', type=int, default=100,
                        help = 'Minimum number of GRID groups required, below which data will be omitted if task us subset_grid_len.')
    parser.add_argument('--prop_filter', nargs=3, metavar=('KEY', 'MIN', 'MAX'), default=['elasticity.K_VRH', '0', 'inf'],
                        help = 'Property (e.g. `elasticity.K_VRH`) contained in data_source, and the min/max values it can take.')

    args = parser.parse_args()

    comp_dict = {'type': args.elem_method, 'elem': args.elem_list}
    
    data_prop = []
    data_prop.append(args.prop_filter[0])
    for val in args.prop_filter[1:]:
        val = val.strip("'").strip('"')
        if 'inf' in val:
            if val[0] == '-':
                data_prop.append(-np.inf)
            else:
                data_prop.append(np.inf)
        else:
            data_prop.append(float(val))
        
    spacegroups = []
    for sg in args.spacegroups:
        if '-' in sg:
            start = int(sg.split('-')[0])
            end = int(sg.split('-')[1])
            spacegroups += range(start, end+1, 1)
        else:
            spacegroups.append(int(sg))
            
    
    
    data = main(data_source = args.data_source,
                tasks = args.tasks,
                composition = comp_dict,
                output_file = args.output_file,
                output_dir = args.output_dir,
                max_dist = args.max_dist,  
                num_grid_shells = args.min_grid_groups,
                data_property = data_prop,
                )
                
    print('There are {} items in data'.format(len(data)))


import unittest
import os
import glob

from gridrdf import data_prepare

from tqdm import tqdm
from functools import partialmethod

# Hack to turn off tqdm output during tests
tqdm.__init__ = partialmethod(tqdm.__init__, disable=True)

FIXTURES_LOC = os.path.join(os.path.dirname(__file__), 'fixtures')

class TestNaCl(unittest.TestCase):
    def setUp(self):
        self.nacl = data_prepare.nacl()
    def test_type(self):
        self.assertIsInstance(self.nacl, data_prepare.Structure)
    def test_lattice(self):
        self.assertEqual(self.nacl.lattice.a, 5.6)
        self.assertTrue(self.nacl.lattice.is_orthogonal)
    def test_atoms(self):
        self.assertEqual(len(self.nacl), 8)
        
class TestDistortedPerovskite(unittest.TestCase):
    def setUp(self):
        self.pero = data_prepare.make_distorted_perovskite(outfile=None)
    def test_length(self):
        self.assertEqual(len(self.pero), 6)
    def test_keys(self):
        self.assertIn('cif', self.pero[0])
        
class TestPerovskiteLattice(unittest.TestCase):
    def setUp(self):
        self.pero = data_prepare.perovskite_different_lattice(outfile=None)
    def test_length(self):
        self.assertEqual(len(self.pero), 61)
        
        
class TestBatchRDF(unittest.TestCase):
    """ Test that batch_RDF correctly produces files"""
    def setUp(self):
        os.mkdir('./_test_fixtures')
        self.data = data_prepare.make_distorted_perovskite(outfile=None)
        data_prepare.batch_rdf(self.data, output_dir = './_test_fixtures')
        self.files = glob.glob('./_test_fixtures/*')
    def test_count(self):
        self.assertEqual(len(self.files), len(self.data))
    def tearDown(self):
        for file in self.files:
            os.remove(file)
        os.rmdir('./_test_fixtures')
        
        
class TestMainOnline(unittest.TestCase):
    """ Check operation of main routine using online perovskite generation. """
    def setUp(self):
        self.data = data_prepare.make_distorted_perovskite(outfile=None)
    def test_composition(self):
        data = data_prepare.main(data_source = 'perovskite_distort',
                          tasks = ['subset_composition',],
                          composition = {'elem':['Ti'], 'type': 'exclude'},
                          output_file = None,
                          )
        self.assertEqual(len(data), 0)
    def test_spacegroup(self):
        data = data_prepare.main(data_source = 'perovskite_distort',
                          tasks = ['subset_space_group',],
                          space_groups = range(1,11,1),
                          output_file = None,
                          )
                          
        # Exact symmetry depends on PyMatGen tolerances, but 
        # cubic should be filtered away.
        self.assertTrue(len(data) < len(self.data))  

    def test_grid_missing(self):
        """ Test correct operation when GRIDS have not been calculated """
        
        with self.assertRaises(OSError):
            data = data_prepare.main(data_source = 'perovskite_distort',
                                     tasks = ['subset_grid_len',],
                                     output_dir = './_test_fixtures',
                                     num_grid_shells = 100)

    def test_grid_bin(self):
        data = data_prepare.main(data_source = 'perovskite_distort',
                          tasks = ['grid_rdf_bin',],
                          output_file = None,
                          output_dir = './_test_fixtures',
                          max_dist = 10,
                          )
                          
        self.assertEqual(len(data), len(glob.glob('./_test_fixtures/*')))
                                     
    def test_grid_kde(self):
        data = data_prepare.main(data_source = 'perovskite_distort',
                          tasks = ['grid_rdf_kde',],
                          output_file = None,
                          output_dir = './_test_fixtures',
                          max_dist = 10,
                          )
                          
        self.assertEqual(len(data), len(glob.glob('./_test_fixtures/*')))
        
    def test_basic_rdf(self):
        data = data_prepare.main(data_source = 'perovskite_distort',
                          tasks = ['original_rdf',],
                          output_file = None,
                          output_dir = './_test_fixtures',
                          max_dist = 10,
                          )
                          
        self.assertEqual(len(data), len(glob.glob('./_test_fixtures/*')))
                  
    def test_multitask(self):  
        data = data_prepare.main(data_source = 'perovskite_distort',
                          tasks = ['grid_rdf_kde', 'subset_grid_len'],
                          output_file = None,
                          output_dir = './_test_fixtures',
                          max_dist = 6,
                          num_grid_shells = 80
                          )
                          
        self.assertEqual(len(data), len(glob.glob('./_test_fixtures/*')))                                               
                                     
    def tearDown(self):
        if os.path.isdir('./_test_fixtures'):
            for file in glob.glob('./_test_fixtures/*'):
                os.remove(file)
            os.rmdir('./_test_fixtures')
        
class TestMainOffline(unittest.TestCase):
    """ Check operation of main routine using test files """
    def setUp(self):
        self.pero_json = os.path.join(FIXTURES_LOC, 'pero_lattice.json')
        self.pero_grids = os.path.join(FIXTURES_LOC, 'pero_grids')
        self.orig_len = 61
    def test_rdf_len(self):
        data = data_prepare.main(data_source = self.pero_json,
                          tasks = ['subset_grid_len',],
                          output_dir = self.pero_grids,
                          num_grid_shells = 100,
                          )
        self.assertEqual(len(data), 13)

        
if __name__ == '__main__':
    unittest.main()


'''
Other types of RDF as benchmark to the GRID.
'''


import numpy as np
import os
try:
    from pymatgen import Structure
except ImportError:
    from pymatgen.core.structure import Structure   
    
try:
    from matminer.featurizers.structure import RadialDistributionFunction
    from matminer.featurizers.structure import PartialRadialDistributionFunction
except:
    print('matminer is not installed, cannot calculate original RDF')


def origin_rdf_histo(data, max_dist=10, bin_size=0.1, output_dir='./'):
    '''
    Calcualte the vanilla RDF using matminer.  
    BEWARE! In the implementation of RDF in matminer, the endpoint
    is not included, e.g. with the default max_dist and bin_size, the  
    last x value is 9.9 instead of 10.0, this makes the list one item  
    shorter than the extend RDF.
    
    Args:
        data: input data from Materials Project
    Return:
        The RDFs are saved into files
    '''
    if not os.path.isdir(output_dir):
        os.mkdir(output_dir)
   
    rdf_fn = RadialDistributionFunction(cutoff=max_dist, bin_size=bin_size)
    for d in data:
        struct = Structure.from_str(d['cif'], fmt='cif')
        rdf_bin = rdf_fn.featurize(struct)
        outfile = os.path.normpath(os.path.join(output_dir, d['task_id']))
        np.savetxt(outfile, rdf_bin, delimiter=' ', fmt='%.3f')
    return


def partial_rdf(data, max_dist=10, bin_size=0.1):
    '''
    Partial RDF (implemented in matminer)

    Schütt, K. T., et al. (2014). 
    How to represent crystal structures for machine learning:  
    Towards fast prediction of electronic properties. 
    Physical Review B 89(20).205118
	
    Args:
        data: input data from Materials Project
    Return:
        The RDFs are saved into files
    '''
    prdf_fn = PartialRadialDistributionFunction(cutoff=max_dist, bin_size=bin_size)

    structs = []
    ids = []
    for d in data:
        structs.append(Structure.from_str(d['cif'], fmt='cif'))
        ids.append(d['task_id'])
    prdf_fn.fit(structs)

    for i, struct in enumerate(structs):
        prdf_bin = prdf_fn.featurize(struct)
        np.savetxt(ids[i], prdf_bin, delimiter=' ', fmt='%.3f')
    return


def mbtr_rdf():
    '''
    MBTR with k=2 (implemented in describ)

    Huo, H. and M. Rupp (2017) 
    Unified Representation of Molecules and Crystals for Machine Learning. 
    arXiv:1704.06439 

    Args:

    Return:

    '''
    pass





import unittest
import os
import glob

from gridrdf import extendRDF
from gridrdf import data_prepare

from pymatgen.core.structure import Structure

from tqdm import tqdm
from functools import partialmethod

# Hack to turn off tqdm output during tests
tqdm.__init__ = partialmethod(tqdm.__init__, disable=True)

FIXTURES_LOC = os.path.join(os.path.dirname(__file__), 'fixtures')

class TestPairwiseDistances(unittest.TestCase):
    def setUp(self):
        self.nacl = data_prepare.nacl()
        self.low_symm = Structure.from_file(os.path.join(FIXTURES_LOC, 'Hf5CuSn3.cif'))
        self.frac_occ = Structure([[4,0,0],[0,4,0],[0,0,4]], [{'Fe':0.4, 'Mn':0.5}], [[0,0,0]])

    def test_num_sites(self):
        """ Check distance and num_neighbour cutoffs give the same answer in NaCl. """
        neighbours_cutoff = extendRDF.get_pairwise_distances(self.nacl, cutoff = 3)
        neighbours_count = extendRDF.get_pairwise_distances(self.nacl, num_neighbours=6)
        self.assertEqual(len(neighbours_cutoff), 8)
        self.assertEqual(len(neighbours_count), 8)
        self.assertCountEqual(neighbours_count[0], neighbours_cutoff[0])
        self.assertListEqual(neighbours_cutoff, neighbours_count)
    def test_sorted_distances(self):
        """ Check neighbours are sorted by increasing distance. """
        neighbours = extendRDF._sorted_neighbours(self.nacl, 3.0)
        self.assertEqual(len(neighbours), 8)
        for site in neighbours:
            current_dist = 0
            for n in site:
                self.assertGreaterEqual(n.nn_distance, current_dist)
                current_dist = n.nn_distance
    def test_neighbour_cutoffs(self):
        """ Check cutoffs are adequate given a simple geometry.
        
        Note that these values are likely to be most inaccurate for small cutoffs/neighbours, but the 
        returned values should be large enough to work correctly.

        """

        # Check distances give sensible numbers of neighbours
        self.assertGreaterEqual(extendRDF._estimate_cutoff(self.nacl, 6), self.nacl.lattice.a/2)
        self.assertGreaterEqual(extendRDF._estimate_cutoff(self.nacl, 26), self.nacl.lattice.a)

        # Check number of predicted neighbours for a given cutoff
        self.assertLessEqual(extendRDF._estimate_neighbours(self.nacl, self.nacl.lattice.a/2), 6)

        # Check that `get_pairwise_distances` correctly updates cutoff
        neighbours, new_cutoff = extendRDF.get_pairwise_distances(self.nacl, num_neighbours=26, return_cutoff=True)
        self.assertLessEqual(new_cutoff, self.nacl.lattice.a *1.0009)

    def test_neighour_finding(self):
        """ Check that neighbours are found correctly. """

        # Specifying num_neighbours and cutoff together should raise an error
        self.assertRaises(AssertionError, extendRDF.get_pairwise_distances, self.nacl, num_neighbours=50, cutoff=5)

        # Neighbours should have the correct shape
        neighbours = extendRDF.get_pairwise_distances(self.nacl, num_neighbours=50)
        self.assertEqual(len(neighbours), len(self.nacl))
        self.assertListEqual([len(i) for i in neighbours], [50]*len(self.nacl))

        pymatgen_neighbours = self.low_symm.get_all_neighbors(3.5)
        pymatgen_neighbour_count = [len(i) for i in pymatgen_neighbours]
        low_symm_neighbours = extendRDF.get_pairwise_distances(self.low_symm, cutoff=3.5)
        
        # A fixed cutoff should return the same result as PyMatGen
        self.assertListEqual(pymatgen_neighbour_count, [len(j) for j in low_symm_neighbours])

        # A fixed number of neighbours should give equal sizes, but an increased cutoff
        fixed_neighbours = max(pymatgen_neighbour_count)
        low_symm_neighbours, cutoff = extendRDF.get_pairwise_distances(self.low_symm, num_neighbours = fixed_neighbours, return_cutoff=True)
        self.assertListEqual([len(i) for i in low_symm_neighbours], [fixed_neighbours]*len(self.low_symm))

        self.assertGreaterEqual(cutoff, 3.5)

    def test_occupancies(self):
        """ Check occupancies are extracted correctly. """

        neighbours = self.frac_occ.get_all_neighbors(30)

        dists, occs = extendRDF._neighbours_to_dists(neighbours)

        self.assertListEqual(list(dists.shape), list(occs.shape))
        # Check all elements of the partially occupied structure have total 0.9
        self.assertEqual(occs.min(), 0.9)
        self.assertEqual(occs.max(), 0.9)



if __name__ == '__main__':
    unittest.main()


import unittest
import os
import glob
import numpy as np

import gridrdf

from pymatgen.core.structure import Structure

from tqdm import tqdm
from functools import partialmethod

# Hack to turn off tqdm output during tests
tqdm.__init__ = partialmethod(tqdm.__init__, disable=True)

FIXTURES_LOC = os.path.join(os.path.dirname(__file__), 'fixtures')

class TestComposition(unittest.TestCase):
    def setUp(self):
        self.nacl = gridrdf.data_prepare.nacl()
        self.low_symm = Structure.from_file(os.path.join(FIXTURES_LOC, 'Hf5CuSn3.cif'))
        self.frac_occ = Structure([[4,0,0],[0,4,0],[0,0,4]], [{'Fe':0.4, 'Mn':0.5}], [[0,0,0]])
        self.deuterated = Structure(4*np.eye(3), ['D','O'], [[0,0,0],[0.5,0.5,0.5]])

    def test_deuterium(self):
        """ Check that composition works correctly for deuterium. """

        neighbours = gridrdf.extendRDF.get_pairwise_distances(self.deuterated, num_neighbours=10)
        comp = gridrdf.composition.composition_hist(self.deuterated, neighbours=neighbours)
        
        self.assertEqual(comp.sum(axis=0)[1], comp.shape[0] / 2)


if __name__ == '__main__':
    unittest.main()




import numpy as np
from sklearn.linear_model import LinearRegression
from sklearn.metrics import mean_absolute_error, r2_score


data = np.loadtxt('data',delimiter=',').transpose()
X_na = data[0].reshape(-1,1)
X_cp = data[1].reshape(-1,1)
X_both = data[:2,].transpose()
y = data[2]

reg_cp = LinearRegression().fit(X_cp, y)
reg_cp.score(X_cp, y)
mean_absolute_error(reg_cp.predict(X_cp), y)

reg_na = LinearRegression().fit(X_na, y)
reg_na.score(X_na, y)
mean_absolute_error(reg_na.predict(X_na), y)

reg_both = LinearRegression().fit(X_both, y)
reg_both.score(X_both, y)
mean_absolute_error(reg_both.predict(X_both), y)




"""
An example of how to compute GRID distances directly without writing intermediate files.

This example uses manually created PyMatGen Structure objects, but could be modified to 
compute Structures from another source.

NOTE: If the number of Structures gets too large, this could have memory implications.

"""

__author__ = "James Cumby"
__email__ = "james.cumby@ed.ac.uk"

import gridrdf
from pymatgen.core.lattice import Lattice
from pymatgen.core.structure import Structure
import os
import tqdm
import numpy as np
import pandas as pd

script_loc = os.path.dirname(__file__)
data_source_loc = os.path.join(script_loc, '../data_sources')

# Variables to define cutoffs etc
maximum_grid_distance = 10
bin_size = 0.1
broadening = 0.1
number_of_shells = 100

# Set up dummy pymatgen Structures with different cells/compositions
dummy_structures = [
    Structure(Lattice.from_parameters(a=3.84, b=3.84, c=3.84, alpha=120, beta=90, gamma=60), ["Si", "Si"], [[0, 0, 0], [0.75, 0.5, 0.75]]),
    Structure(Lattice.from_parameters(a=3.84, b=3.84, c=3.84, alpha=120, beta=90, gamma=60), ["Ni", "Ni"], [[0, 0, 0], [0.75, 0.5, 0.75]]),
    Structure(Lattice.from_parameters(a=3.94, b=3.94, c=3.94, alpha=120, beta=90, gamma=60), ["Si", "Si"], [[0, 0, 0], [0.75, 0.5, 0.75]]),
]


# First (and slowest), find the ordered distances to neighbouring atoms.
# Note that we specify a certain number of neighbours (number_of_shells) to simplify processing,
# but a distance cutoff could be used instead.
neighbours, cutoffs = gridrdf.extendRDF.find_all_neighbours(dummy_structures,
                                                            num_neighbours=number_of_shells,
                                                            cutoff=None,
                                                            return_limits = True,
                                                            dryrun = False)
    
# Adjust the cutoff so that all nearest neighbours can fit
max_dist = max(maximum_grid_distance, np.round(cutoffs[1], 1))
if max_dist != maximum_grid_distance:
    print(f"Maximum distance has been updated to {max_dist} to account for {number_of_shells} neighbours")

# Empty list to hold GRID arrays
grid_representations = []

# Iteratively calculate GRIDs

# For memory-intensive processing (i.e. large numbers of structures)
# `return_sparse` can be used to reduce the size of each GRID representation (as a sparse array)
for i, struct in tqdm.tqdm(enumerate(dummy_structures)):
    grid_representations.append(gridrdf.extendRDF.calculate_rdf(struct,
                                                                neighbours[i],
                                                                rdf_type='grid',
                                                                max_dist=max_dist,
                                                                bin_width=bin_size,
                                                                smearing = broadening,
                                                                normed = True,
                                                                broadening_method = 'convolve',
                                                                return_sparse = False)
                                )


    
# Calculate EMD similarity between grids using a numba-optimised EMD calculation

# If the 2D dissimilarity matrix is large, `results_array` can be used to store results efficiently,
# for instance by passing a h5py dataset object in order to write straight to disk.

# Note: This is very slow for this example due to numba compilation, but is 
# optimised for larger numbers of structures
grid_similarity = gridrdf.earth_mover_distance.super_fast_EMD_matrix(grid_representations,
                                                                     bin_width = bin_size,
                                                                    )    

grid_similarity = pd.DataFrame(grid_similarity)



# Currently, composition_one_hot requires a list of dicts, each with a 'task_id' for each structure
# NOTE - this is likely to change in a future release
structure_ids = [ {'task_id': i, 'structure':dummy_structures[i]} for i in range(len(grid_representations)) ]

# First, convert composition to vector encoding
elem_vectors, elem_symbols = gridrdf.composition.composition_one_hot(structure_ids, method='percentage')

# Now computed EMD similarity based on "distances" between species contained in `similarity_matrix.csv`)
# This is essentially Pettifor distance, but with non-integer steps defined by data-mining of probabilities.
comp_similarity = gridrdf.earth_mover_distance.composition_similarity_matrix(elem_vectors, 
                                                                     elem_similarity_file = os.path.join(data_source_loc, 'similarity_matrix.csv')
                                                                     )
                                                                     
comp_similarity = comp_similarity.add(comp_similarity.T, fill_value = 0)

total_similarity = 10*grid_similarity + comp_similarity

                                                                     
print('\nStructural Similarity:')
print(grid_similarity)

print('\nComposition Similarity:')
print(comp_similarity)

print('\nTotal similarity (= 10*GRID + Composition):')
print(total_similarity)

                                                                     

